import asyncio
from abc import ABC
from typing import Any

import orjson

from integra_bridge.adapters.base_input_connector import BaseInputConnectorAdapter
from integra_bridge.dto import Exchange, ConnectorToBlockView
from integra_bridge.dto.body import Body
from integra_bridge.dto.redeploy import ConnectionParams


class SingleModeInputConnectorAdapter(BaseInputConnectorAdapter, ABC):
    """
    Данный простой тип коннекторов предназначен для работы в однопоточной среде. Данные о всех зарегистрированных
    инстансах хранятся в оперативной памяти. Не подойдет при работе в кластере или использовании нескольких воркеров.
    """

    _connector_to_block_views: dict[str, ConnectorToBlockView] = {}

    @classmethod
    async def push_to_integra(cls, input_body: Any, connect_to_block_id: str, *args: Any, **kwargs: Any) -> int:
        try:
            string_body = orjson.dumps(input_body).decode(encoding="utf-8")
            body_type = "json"
        except Exception:
            string_body = input_body
            body_type = "string"

        connector_to_block_view = SingleModeInputConnectorAdapter._connector_to_block_views.get(connect_to_block_id)
        if not connector_to_block_view:
            raise ValueError(f"connect to block view {connect_to_block_id} not found")

        input_body = Body(stringBody=string_body, type=body_type)
        exchange = Exchange(inputBody=input_body)
        exchange.block_id = connector_to_block_view.block_id
        exchange.input_connect_id = connector_to_block_view.connect_id
        exchange.company_id = connector_to_block_view.company_id

        connector_url = f"{connector_to_block_view.url_integra_service}/api/external/connector/input/{connect_to_block_id}?connectorTitle={connector_to_block_view.connector_title}"
        return await cls._send(exchange, connector_url)

    async def deploy_input_flow(self, exchange: Exchange, connector_title: str, *args: Any, **kwargs: Any):
        connector_to_block_view = ConnectorToBlockView(
            connector_title=connector_title,
            company_id=exchange.company_id,
            block_id=exchange.block_id,
            connect_id=exchange.input_connect_id,
            url_integra_service=exchange.headers.get('urlIntegra'),
            exchange_deploy=exchange
        )
        connector_to_block_id = await connector_to_block_view.get_id()
        connection_params = exchange.input_connect.params or {}
        SingleModeInputConnectorAdapter._connector_to_block_views[connector_to_block_id] = connector_to_block_view
        print('!!!!!!!!DEPLOYED:',
              SingleModeInputConnectorAdapter._connector_to_block_views[connector_to_block_id].connector_title,
              SingleModeInputConnectorAdapter._connector_to_block_views.keys())
        asyncio.create_task(self.on_after_deploy(
            connection_id=connector_to_block_id,
            connection_params=connection_params
        ))
        return exchange

    async def redeploy_input_flow(self, connection_params: ConnectionParams, connector_title: str, *args: Any,
                                  **kwargs: Any):
        connector_to_block_view = ConnectorToBlockView(
            connector_title=connector_title,
            company_id=connection_params.company_id,
            block_id=connection_params.block_id,
            connect_id=connection_params.connect_id,
            url_integra_service=connection_params.url_integra_service,
        )
        connector_to_block_id = await connector_to_block_view.get_id()
        SingleModeInputConnectorAdapter._connector_to_block_views[connector_to_block_id] = connector_to_block_view
        print('!!!!!!!!REDEPLOYED:',
              SingleModeInputConnectorAdapter._connector_to_block_views[connector_to_block_id].connector_title,
              SingleModeInputConnectorAdapter._connector_to_block_views.keys())
        asyncio.create_task(self.on_after_redeploy(
            connection_id=connector_to_block_id,
            connection_params=connection_params.params
        ))

    async def destroy_input_flow(self, exchange: Exchange, connector_title: str, *args: Any, **kwargs: Any):
        connector_to_block_view = ConnectorToBlockView(
            connector_title=connector_title,
            company_id=exchange.company_id,
            block_id=exchange.block_id,
            connect_id=exchange.input_connect_id,
            exchange_deploy=exchange
        )
        connector_to_block_id = await connector_to_block_view.get_id()
        SingleModeInputConnectorAdapter._connector_to_block_views.pop(connector_to_block_id, None)
        print('!!!!!!!!DESTROYED:', SingleModeInputConnectorAdapter._connector_to_block_views.keys())
        asyncio.create_task(self.on_after_destroy(
            connection_id=connector_to_block_id,
            connection_params=connector_to_block_view.model_dump()
        ))
        return exchange



import asyncio
from abc import ABC
from typing import Any

import orjson

from integra_bridge.adapters.base_input_connector import BaseInputConnectorAdapter
from integra_bridge.dto import Exchange, ConnectorToBlockView
from integra_bridge.dto.body import Body
from integra_bridge.dto.redeploy import ConnectionParams


class MultiprocessInputConnectorAdapter(BaseInputConnectorAdapter, ABC):
    """
    Данный тип коннекторов предназначен для работы в многопоточной среде. Данные о всех зарегистрированных инстансах
    перенаправляются методами on_after_deploy и on_after_redeploy для описания логики хранения и обработки с
    использованием СУБД.
    """

    @classmethod
    async def push_to_integra(cls, input_body: Any, connect_to_block_id: str, *args: Any, **kwargs: Any):
        url_integra_service = kwargs.get('url_integra_service')
        if not url_integra_service:
            raise Exception('url_integra_service is not defined')

        try:
            string_body = orjson.dumps(input_body).decode(encoding="utf-8")
            body_type = "json"
        except Exception:
            string_body = input_body
            body_type = "string"

        input_body = Body(stringBody=string_body, type=body_type)
        exchange = Exchange(inputBody=input_body)

        try:
            company_id, block_id, connect_id = connect_to_block_id.split('_')
            exchange.block_id = block_id
            exchange.input_connect_id = connect_id
            exchange.company_id = company_id
        except Exception:
            raise Exception(f'connect_to_block_id have wrong format: {connect_to_block_id}')

        return await cls._send(exchange, url_integra_service)

    async def deploy_input_flow(self, exchange: Exchange, connector_title: str, *args: Any, **kwargs: Any) -> int:
        connector_to_block_view = ConnectorToBlockView(
            connector_title=connector_title,
            company_id=exchange.company_id,
            block_id=exchange.block_id,
            connect_id=exchange.input_connect_id,
            url_integra_service=exchange.headers.get('urlIntegra'),
            exchange_deploy=exchange
        )
        connector_to_block_id = await connector_to_block_view.get_id()
        connector_url = f"{connector_to_block_view.url_integra_service}/api/external/connector/input/{connector_to_block_id}?connectorTitle={connector_to_block_view.connector_title}"
        connection_params = exchange.input_connect.params or {}

        # ВРЕМЕННАЯ ЗАГЛУШКА
        connection_auth = {}

        print('!!!!!!!!DEPLOYED:', connector_to_block_view)

        asyncio.create_task(self.on_after_deploy(
            connection_id=connector_to_block_id,
            connection_params=connection_params,
            connector_url=connector_url,
            auth=connection_auth
        ))
        return exchange

    async def redeploy_input_flow(self, connection_params: ConnectionParams, connector_title: str, *args: Any,
                                  **kwargs: Any):
        connector_to_block_view = ConnectorToBlockView(
            connector_title=connector_title,
            company_id=connection_params.company_id,
            block_id=connection_params.block_id,
            connect_id=connection_params.connect_id,
            url_integra_service=connection_params.url_integra_service,
        )
        connector_to_block_id = await connector_to_block_view.get_id()
        connector_url = f"{connector_to_block_view.url_integra_service}/api/external/connector/input/{connector_to_block_id}?connectorTitle={connector_to_block_view.connector_title}"
        print('!!!!!!!!REDEPLOYED:', connector_to_block_view)
        asyncio.create_task(self.on_after_redeploy(
            connection_id=connector_to_block_id,
            connection_params=connection_params.params,
            connector_url=connector_url,
            auth=connection_params.auth
        ))

    async def destroy_input_flow(self, exchange: Exchange, connector_title: str, *args: Any, **kwargs: Any):
        connector_to_block_view = ConnectorToBlockView(
            connector_title=connector_title,
            company_id=exchange.company_id,
            block_id=exchange.block_id,
            connect_id=exchange.input_connect_id,
            exchange_deploy=exchange
        )
        connector_to_block_id = await connector_to_block_view.get_id()
        print('!!!!!!!!DEPLOYED:', connector_to_block_view)
        asyncio.create_task(self.on_after_destroy(
            connection_id=connector_to_block_id,
            connection_params=connector_to_block_view.model_dump()
        ))
        return exchange



from pydantic import BaseModel, ConfigDict, StrictStr, Field


class ConnectionParams(BaseModel):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    company_id: StrictStr | None = Field(None, alias="companyId")
    block_id: StrictStr | None = Field(None, alias="blockId")
    connect_id: StrictStr | None = Field(None, alias="connectId")
    url_integra_service: StrictStr | None = Field(None, alias="urlIntegraService")
    params: dict = Field(default_factory=dict, alias="params")
    auth: dict = Field(default_factory=dict, alias="auth")


class RedeployConnectors(BaseModel):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    input_connectors: dict[str, list[ConnectionParams]] = Field(default_factory=dict, alias="inputConnectors")



from pydantic import BaseModel


class ProcessedComponents(BaseModel):
    body: dict | list | str
    input_headers: dict
    headers: dict
    context: dict | list | str | None = None
    params: dict | None = None



from abc import ABC

from integra_bridge.adapters.base_adapter import BaseAdapter
from integra_bridge.common.enums import AdapterType

from integra_bridge.dto.responces.validation import ValidationResponse
from integra_bridge.entity.connector import Connector
from integra_bridge.entity.output_status import OutputStatus


class OutputConnectorAdapter(BaseAdapter, ABC):

    def __init__(self):
        super().__init__()
        self.__view = None
        BaseAdapter.add_adapter(self, adapter_type=AdapterType.connectors)

    def __del__(self):
        BaseAdapter.remove_adapter(self, adapter_type=AdapterType.connectors)

    async def pull_from_integra(self, input_body: dict, params: dict, headers: dict) -> OutputStatus:
        status = OutputStatus()
        return status

    async def validate_output(self, connector: Connector) -> ValidationResponse:
        return ValidationResponse(result=True)



from starlette import status
from integra_bridge.api_router import APIRouter

from fastapi import Response

service_router = APIRouter(prefix='/service', tags=["Обслуживание сервиса"])


@service_router.get('/checkhealth')
async def checkhealth():
    return Response(status_code=status.HTTP_200_OK)



from abc import ABC, abstractmethod

from integra_bridge.common.enums import AdapterType
import uuid


class BaseAdapter(ABC):
    _adapters: dict[str, set] = dict()

    def __init__(self):
        self.__view = None
        self.__uid = uuid.uuid4()

    @abstractmethod
    async def get_view(self):
        ...

    @property
    def view(self):
        return self.__view

    @classmethod
    def get_adapters(cls, adapter_type: AdapterType) -> set:
        return cls._adapters.get(adapter_type.value, set())

    @classmethod
    def add_adapter(cls, service, adapter_type: AdapterType):
        group = cls._adapters.get(adapter_type.value, None)
        if group:
            group.add(service)
        else:
            cls._adapters[adapter_type.value] = {service, }

    @classmethod
    def remove_adapter(cls, service, adapter_type: AdapterType):
        group = cls._adapters.get(adapter_type.value, None)
        if group:
            group.remove(service)

    def __hash__(self):
        return hash(self.__uid)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__uid == other.__uid



from abc import ABC, abstractmethod
from typing import Any

import httpx
import orjson
from starlette import status

from integra_bridge.dto.responces.validation import ValidationResponse
from integra_bridge.adapters.base_adapter import BaseAdapter
from integra_bridge.common.enums import AdapterType
from integra_bridge.common.settings import SETTINGS
from integra_bridge.dto import Exchange
from integra_bridge.entity.connector import Connector


class BaseInputConnectorAdapter(BaseAdapter, ABC):

    def __init__(self):
        super().__init__()
        self.__view = None
        BaseAdapter.add_adapter(self, adapter_type=AdapterType.connectors)

    def __del__(self):
        BaseAdapter.remove_adapter(self, adapter_type=AdapterType.connectors)

    @abstractmethod
    async def push_to_integra(self, input_body: Any, connect_to_block_id: str, *args: Any, **kwargs: Any):
        ...

    @abstractmethod
    async def deploy_input_flow(self, exchange: Exchange, connector_title: str, *args: Any, **kwargs: Any):
        ...

    @abstractmethod
    async def destroy_input_flow(self, exchange: Exchange, connector_title: str, *args: Any, **kwargs: Any):
        ...

    @abstractmethod
    async def on_after_deploy(self, connection_id: str, connection_params: dict, connector_url: str | None = None,
                              auth: dict | None = None, *args: Any, **kwargs: Any) -> None:
        ...

    async def on_after_redeploy(self, connection_id: str, connection_params: dict, connector_url: str | None = None,
                                auth: dict | None = None, *args: Any, **kwargs: Any) -> None:
        ...

    @abstractmethod
    async def on_after_destroy(self, connection_id: str, connection_params: dict, *args: Any, **kwargs: Any) -> None:
        ...

    async def redeploy_input_flow(self, connection_params: dict, connector_title: str, *args: Any, **kwargs: Any):
        ...

    async def validate_input(self, connector: Connector) -> ValidationResponse:
        return ValidationResponse(result=True)

    @classmethod
    async def _send(cls, exchange: Exchange, connector_url: str) -> int:
        auth = httpx.BasicAuth(username='admin', password='admin')
        async with httpx.AsyncClient(auth=auth, timeout=SETTINGS.DEFAULT_CONNECTOR_TIMEOUT) as client:
            try:
                response = await client.post(
                    connector_url,
                    content=orjson.dumps(exchange.model_dump(by_alias=True)),
                    headers={"Content-Type": "application/json"},
                    timeout=SETTINGS.DEFAULT_CONNECTOR_TIMEOUT
                )
            except httpx.TimeoutException:
                return status.HTTP_504_GATEWAY_TIMEOUT
            except ValueError as err:
                print('Error while sending to integra: ', str(err))
                return status.HTTP_406_NOT_ACCEPTABLE
            return response.status_code



import traceback

import orjson
import xmltodict
from pydantic import ValidationError

from integra_bridge.adapters import ProcessorAdapter
from fastapi import HTTPException, Request
from starlette import status

from integra_bridge.dto import Exchange
from integra_bridge.dto.responces.validation import ValidationResponse
from integra_bridge.entity.processed_components import ProcessedComponents
from integra_bridge.entity.processor import Processor


class ProcessorHandler:
    @classmethod
    async def execute(cls, request: Request, processor_title: str) -> Exchange:
        processor_adapter = await cls.__get_processor_by_title(processor_title)

        if isinstance(request, Request):
            # Дисериализация входных данных
            try:
                exchange = await request.json()
            except Exception:
                raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST,
                                    detail="Input body is not JSON serializable")
        else:
            exchange = request

        # Дисериализация exchange в python объекты
        exchange_data = await ProcessorHandler.exchange_deserialization(exchange)

        processed_components = ProcessedComponents(
            body=exchange_data['inputBody'],
            input_headers=exchange_data['inputHeaders'],
            headers=exchange_data['headers'],
            context=exchange_data['context'],
            params=exchange_data['params'],
        )
        try:
            processed_exchange_components = await processor_adapter.execute(processed_components)
        except Exception as e:
            exception = traceback.format_exc()
            exception_lines = exception.splitlines()[3:]
            formatted_exception = "\n".join(exception_lines).strip()
            formatted_exception = formatted_exception[2:]
            exchange["exception"] = str(e)
            exchange['stackTrace'] = formatted_exception
        else:
            context_exists = True if exchange_data['context_exists'] or processed_exchange_components.context else False
            processed_exchange = await ProcessorHandler.exchange_serialization(
                exchange=exchange,
                context_exists=context_exists,
                body=processed_exchange_components.body,
                context=processed_exchange_components.context,
                input_headers=processed_exchange_components.input_headers,
                headers=processed_exchange_components.headers,
            )
            try:
                exchange = Exchange(**processed_exchange)
            except ValidationError as e:
                raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=str(e))
        return exchange

    @classmethod
    async def validate(cls, processor: Processor, title: str) -> ValidationResponse:
        try:
            processor_adapter = await cls.__get_processor_by_title(title)
        except Exception as e:
            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=str(e))
        return await processor_adapter.validate(processor)

    @classmethod
    async def __get_processor_by_title(cls, title: str) -> ProcessorAdapter:
        from integra_bridge.common.enums import AdapterType
        for processor_adapter in ProcessorAdapter.get_adapters(adapter_type=AdapterType.processors):
            processor_view = await processor_adapter.get_view()
            if processor_view.title.lower() == title.lower():
                return processor_adapter
        raise HTTPException(status_code=404, detail=f'Processor not found: {title} ')

    @staticmethod
    async def exchange_deserialization(exchange: dict):
        body_str = exchange.get("body", {}).get("stringBody", "{}")
        body_type = exchange.get('body', {}).get('type', 'string')

        # Получам контекст если нет то создаем
        context = exchange.get('context')
        context_str = context.get('stringBody') if context else '{}'
        try:
            input_context = orjson.loads(context_str)
        except ValueError:
            input_context = context_str

        # Получам входящие input headers если нет то создаем
        input_headers = exchange.get('inputHeaders', {})

        # Получам входящие headers если нет то создаем
        headers = exchange.get('headers', {})

        try:
            input_body = orjson.loads(body_str)
        except ValueError:
            input_body = body_str

        # Получаем параметры формы
        params = exchange.get("processor", {}).get("params", {})

        return dict(
            inputBody=input_body,
            inputHeaders=input_headers,
            headers=headers,
            context=input_context,
            context_exists=True if context else False,
            input_body_type=body_type,
            params=params
        )

    @staticmethod
    async def exchange_serialization(
            exchange: dict,
            context_exists: bool,
            body: dict | str,
            context: dict,
            input_headers: dict,
            headers: dict,
    ) -> dict:
        new_body = body
        exchange["inputHeaders"] = input_headers
        exchange["headers"] = headers
        # Сериализация в исходный формат JSON/String
        try:
            new_body_json = orjson.dumps(new_body).decode(encoding="utf-8")
            if isinstance(new_body, dict | list):
                new_body = new_body_json
                body_type = "json"
            else:
                raise Exception
        except Exception:
            try:
                xmltodict.parse(new_body)
                body_type = "xml"
            except Exception:
                body_type = "string"
        exchange["body"]["type"] = body_type
        exchange["body"]["stringBody"] = new_body if body_type == "json" else str(new_body)
        if not context_exists:
            return exchange

        new_context = context
        try:
            context_data_json = orjson.dumps(new_context).decode(encoding="utf-8")
            if isinstance(new_context, dict | list):
                context_data = context_data_json
                context_type = "json"
            else:
                raise Exception
        except Exception:
            context_type = "string"
            context_data = str(new_context)
        exchange["context"] = {
            "type": context_type,
            "stringBody": context_data,
            "outputBody": context_data,
        }
        return exchange



from .base_processor import ProcessorAdapter
from .single_mode_connector import SingleModeInputConnectorAdapter
from .output_connector import OutputConnectorAdapter
from .multiprocess_connector import MultiprocessInputConnectorAdapter


__all__ = (
    'ProcessorAdapter',
    'OutputConnectorAdapter',
    'SingleModeInputConnectorAdapter',
    'MultiprocessInputConnectorAdapter',
)



from fastapi.responses import FileResponse

from integra_bridge.api_router import APIRouter
from integra_bridge.dto.responces.external_service import ExternalServiceConfigResponse
from integra_bridge.common.dependency_manager import dm
from integra_bridge.services.configuration import ConfigurationHandler

configuration_router = APIRouter(prefix='/configuration', tags=["Работа с конфигурациями внешних сервисов"])


@configuration_router.get(
    path='/',
    response_model=ExternalServiceConfigResponse,
    response_model_exclude_none=True,
    response_model_by_alias=True,
)
async def get_configurations() -> ExternalServiceConfigResponse:
    return await ConfigurationHandler.get_configurations()


@configuration_router.get(
    path='/manual'
)
async def get_manual():
    manual_path = dm.manual_path
    if manual_path.exists():
        return FileResponse(manual_path, media_type='application/octet-stream', filename=dm.manual_path.name)
    else:
        return {"error": "File not found"}
